package sketch;

import shared.Tools;
import tax.TaxTree;

public class DisplayParams {
	
	public static DisplayParams parseDoubleHeader(String s){
		if(!s.startsWith("##")){return SketchObject.defaultParams;}
		StringBuilder sb=new StringBuilder();
		for(int i=2; i<s.length(); i++){
			char c=s.charAt(i);
			if(c=='\n'){break;}
			sb.append(c);
		}
		return parseDoubleHeaderLine(sb.toString());
	}
	
	public static DisplayParams parseDoubleHeaderLine(String line) {
		if(line.startsWith("##")){line=line.substring(2);}
		else{assert(!line.startsWith("#")) : line;}
		if(line.length()<1){return SketchObject.defaultParams;}
		
		DisplayParams params=new DisplayParams();
		
		String[] args=line.split(" ");
		for(String arg : args){
			String[] split=arg.split("=");
			String a=split[0].toLowerCase();
			String b=split.length>1 ? split[1] : null;
			if(b==null || b.equalsIgnoreCase("null")){b=null;}
			while(a.startsWith("-")){a=a.substring(1);} //Strip leading hyphens
			
			boolean x=params.parse(arg, a, b);
			assert(x) : "Unknown parameter "+arg+"\n"+line;
		}
		
		return params;
	}
	
	public boolean parse(String arg, String a, String b){
	
		if(a.equals("mincount") || a.equals("minhits")  || a.equals("hits") || a.equals("matches")){
			minCount=Integer.parseInt(b);
		}else if(a.equals("minid") || a.equals("id")){
			minID=Float.parseFloat(b);
			if(minID>1){minID/=100;}
		}else if(a.equals("records") || a.equals("maxrecords")){
			maxRecords=Integer.parseInt(b);
			assert(maxRecords>=1) : "Max records must be at least 1.";
		}else if(a.equals("format")){
			format=Integer.parseInt(b);
		}else if(a.equals("level") || a.equals("taxlevel") || a.equals("minlevel")){
			if(b==null){taxLevel=-1;}
			else if(Character.isDigit(b.charAt(0))){
				taxLevel=Integer.parseInt(b);
			}else{
				taxLevel=TaxTree.stringToLevel(b.toLowerCase());
			}
		}else if(a.equals("printtax") || a.equals("printtaxa")){
			printTax=Tools.parseBoolean(b);
		}else if(a.equals("printoriginalname") || a.equals("printseqname") || a.equals("printname0")){
			printOriginalName=Tools.parseBoolean(b);
		}else if(a.equals("samplerate")){
			samplerate=Float.parseFloat(b);
		}else if(a.equals("mode")){
			mode=SketchObject.parseMode(arg, a, b);
		}else{
			return false;
		}
		return true;
	}
	
	public String toString(){
		StringBuilder sb=new StringBuilder();
		sb.append("##");
		sb.append("hits=").append(minCount);
		sb.append(" id=").append(String.format("%.5f",minID));
		sb.append(" records=").append(maxRecords);
		sb.append(" format=").append(format);
		sb.append(" level=").append(taxLevel);
		sb.append(" printtax=").append(printTax);
		if(mode!=default_mode){sb.append(" mode=").append(mode);}
		if(samplerate!=default_samplerate){sb.append(" samplerate=").append(String.format("%.4f",samplerate));}
		if(printOriginalName!=default_printOriginalName){sb.append(" printname0=").append(printOriginalName);}
		sb.append('\n');
		return sb.toString();
	}
	
	public int maxRecords=default_maxRecords;
	public float minID=default_minID;
	public int format=default_format;
	public boolean printTax=default_printTax;
	public boolean printOriginalName=default_printOriginalName;
	public int minCount=default_minCount;
	public int taxLevel=default_taxLevel;
	public int mode=default_mode;
	public float samplerate=default_samplerate;
	
	public static final int default_maxRecords=100;
	public static final float default_minID=0.0002f;
	public static final int default_format=2;
	public static final boolean default_printTax=false;
	public static final boolean default_printOriginalName=true;
	public static final int default_minCount=3;
	public static final int default_taxLevel=TaxTree.SPECIES;
	public static final int default_mode=SketchObject.ONE_SKETCH;
	public static final float default_samplerate=1;
	
	
	
}
